@@ -1,6 +1,7 @@
 'use strict';
 
-var NQPObject = require('./nqp-object.js');
+const NQPObject = require('./nqp-object.js');
+const NQPStr = require('./nqp-str.js');
 
 class NQPInt extends NQPObject {
   constructor(value) {
@@ -8,8 +9,8 @@ class NQPInt extends NQPObject {
     this.value = value | 0;
   }
 
-  Str(ctx, _NAMED, self) {
-    return this.value.toString();
+  p6$Str(ctx, _NAMED, self) {
+    return new NQPStr(this.value.toString());
   }
 
   $$toBool(ctx) {
